/*
 * rootapps graphics main management functions
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/time.h>
#include "options.h"
#include "rootapp.h"


#define HASAPP if (! app) {fprintf(stderr, "Application not initialized !\n"); exit(1);}




unsigned long get_color(const char *color_name) {
    Colormap      colormap;
    XColor        color;
    int           screen = DefaultScreen(display);
    unsigned long res    = WhitePixel(display, screen);

    colormap = DefaultColormap(display, screen);

    if (color_name) {
        if (! XParseColor(display, colormap, color_name, &color))
            fprintf(stderr, "Cannot find color \"%s\" !\n", color_name);
        else if (! XAllocColor(display, colormap, &color))
            fprintf(stderr, "Cannot allocate color \"%s\" !\n", color_name);
        else {
            res = color.pixel;
        }
    }

    return res;
}


void set_color(unsigned long color) {
    XSetForeground(display, gc, color);
}


void set_color_name(const char *color_name) {
    XSetForeground(display, gc, get_color(color_name));
}


void load_font(const char *font_name) {
    Font font;

    if (! font_name) return;
    font = XLoadFont(display, font_name);
    XSetFont(display, gc, font);
}


int font_height() {
    XFontStruct *fontstruct;
    GContext     ctxt;
    int          res = 0;

    ctxt = XGContextFromGC(gc);
    if ((fontstruct = XQueryFont(display, ctxt)) == NULL) {
        fprintf(stderr, "Cannot get font infos !\n");
        exit(1);
    }
    res = fontstruct->ascent + fontstruct->descent;
    XFreeFontInfo(NULL, fontstruct, 1);
    return res;
}


int font_ascent() {
    XFontStruct *fontstruct;
    GContext     ctxt;
    int          res = 0;

    ctxt = XGContextFromGC(gc);
    if ((fontstruct = XQueryFont(display, ctxt)) == NULL) {
        fprintf(stderr, "Cannot get font infos !\n");
        exit(1);
    }
    res = fontstruct->ascent;
    XFreeFontInfo(NULL, fontstruct, 1);
    return res;
}


int font_descent() {
    XFontStruct *fontstruct;
    GContext     ctxt;
    int          res = 0;

    ctxt = XGContextFromGC(gc);
    if ((fontstruct = XQueryFont(display, ctxt)) == NULL) {
        fprintf(stderr, "Cannot get font infos !\n");
        exit(1);
    }
    res = fontstruct->descent;
    XFreeFontInfo(NULL, fontstruct, 1);
    return res;
}


int text_width(const char *text) {
    XFontStruct *fontstruct;
    GContext     ctxt;
    int          res = 0;

    ctxt = XGContextFromGC(gc);
    if ((fontstruct = XQueryFont(display, ctxt)) == NULL) {
        fprintf(stderr, "Cannot get font infos !\n");
        exit(1);
    }
    res = XTextWidth(fontstruct, text, strlen(text)) + 2;
    XFreeFontInfo(NULL, fontstruct, 1);
    return res;
}


Pixmap copy_pixmap(RootApp *app, Pixmap source) {
    Pixmap   result = None;
    Window   parent;
    unsigned px, py, pw, ph, pb, depth;

    HASAPP;
    if (source == None) {
        fprintf(stderr, "Trying to copy a nonexistant pixmap !\n");
        return None;
    }
    XGetGeometry(display, source, &parent, &px, &py, &pw, &ph, &pb, &depth);
    result = XCreatePixmap(display, app->window, pw, ph, depth);
    XCopyArea(display, source, result, gc, 0, 0, pw, ph, 0, 0);

    return result;
}


Pixmap root_pixmap(const char *desc) {
    Pixmap res = None;
    Atom   id  = None;

    id = XInternAtom(display, desc, True);
    if (id != None) {
        Window         rootwin;
        Atom           act_type;
        int            act_format;
        unsigned long  nitems, bytes_after;
        unsigned char *prop = NULL;

        rootwin = DefaultRootWindow(display);
        if (XGetWindowProperty(display, rootwin, id, 0, 1, False,
                               XA_PIXMAP, &act_type, &act_format,
                               &nitems, &bytes_after, &prop) == Success) {
            if (prop) {
                res = *((Pixmap*) prop);
            }
        }
    }

    return res;
}


Pixmap get_root_pixmap(RootApp *app) {
    Pixmap res;
    
    HASAPP;
    res = root_pixmap("_XROOTPMAP_ID");
    if (res == None) res = root_pixmap("_XSETROOT_ID");

    if (res != None) {
        Pixmap z, winpixmap;
        Window parent;
        int    build = 0;
        int    screen = DefaultScreen(display);
        unsigned int px, py, pw, ph, pb, pd, sw, sh, dx, dy;

        sw = DisplayWidth(display, screen);
        sh = DisplayHeight(display, screen);
        dx = 0;
        dy = 0;
        XGetGeometry(display, res, &parent, &px, &py, &pw, &ph, &pb, &pd);
        if ((pw < sw * 2) || (ph < sh * 2)) {
            z = XCreatePixmap(display, app->window, sw * 2, sh * 2, pd);
            while (dy < sh * 2) {
                while (dx < sw * 2) {
                    int w, h;
                    w = sw*2 - dx > pw ? pw : sw*2 - dx;
                    h = sh*2 - dy > ph ? ph : sh*2 - dy;
                    XCopyArea(display, res, z, gc, 0, 0, w, h, dx, dy);
                    dx += pw;
                }
                dy += ph;
                dx = 0;
            }
            res = z;
            build = 1;
        }
        winpixmap = XCreatePixmap(display, app->window,
                app->wsize.width, app->wsize.height, pd);
        XCopyArea(display, res, winpixmap, gc,
                app->wsize.x < 0 ? app->wsize.x + sw : app->wsize.x,
                app->wsize.y < 0 ? app->wsize.y + sh : app->wsize.y,
                app->wsize.width, app->wsize.height, 0, 0);
        if (build) XFreePixmap(display, res);
        res = winpixmap;
    } else {
        fprintf(stderr, "Could not get root pixmap !\n");
    }
    return res;
}


void pixmap_to_app(RootApp *app, Pixmap source,
        int srcx, int srcy, int width, int height, int destx, int desty) {
    XCopyArea(display, source, app->bgpixmap, gc,
            srcx, srcy, width, height, destx, desty);
}


void draw_line(RootApp *app, int x1, int y1, int x2, int y2) {
    XDrawLine(display, app->bgpixmap, gc,
            app->wsize.x + x1, app->wsize.y + y1,
            app->wsize.x + x2, app->wsize.y + y2);
}


void clean_window(RootApp *app) {
    HASAPP;
    if (app->rootpixmap) {
        XCopyArea(display, app->rootpixmap, app->window, gc,
                0, 0, app->wsize.width, app->wsize.height,
                app->wsize.x, app->wsize.y);
        XFlush(display);
    }
}


void clear_bg(RootApp *app) {
    HASAPP;
    if ((app->rootpixmap) && (app->bgpixmap)) {
        XCopyArea(display, app->rootpixmap, app->bgpixmap, gc,
                0, 0, app->wsize.width, app->wsize.height, 0, 0);
    }
}


void refresh_background(RootApp *app) {
    HASAPP;
    clean_window(app);
    if (app->rootpixmap) XFreePixmap(display, app->rootpixmap);
    if (app->bgpixmap)   XFreePixmap(display, app->bgpixmap);
    app->rootpixmap = get_root_pixmap(app);
    app->bgpixmap = copy_pixmap(app, app->rootpixmap);
}


void display_app(RootApp *app) {
    HASAPP;
    if (app->bgpixmap) {
        XCopyArea(display, app->bgpixmap, app->window, gc,
                0, 0, app->wsize.width, app->wsize.height,
                app->wsize.x, app->wsize.y);
        XFlush(display);
    }
}


static void set_size(RootApp *app, const char *g, int x, int y, int width, int height) {
    int nx, ny, nw, nh, status = 0;
    int sw, sh;
    int screen = DefaultScreen(display);

    HASAPP;
    sw = DisplayWidth(display, screen);
    sh = DisplayHeight(display, screen);

    nx = 0;
    ny = 0;
    if (g) status = XParseGeometry(g, &nx, &ny, &nw, &nh);
    if (!(status & WidthValue)) nw = width;
    if (!(status & HeightValue)) nh = height;
    if (!(status & XValue)) {
        nx = x;
        if (x < 0) nx += sw - nw;
    } else if (status & XNegative) nx += sw - nw;
    if (!(status & YValue)) {
        ny = y;
        if (y < 0) ny += sh - nh;
    } else if (status & YNegative) ny += sh - nh;

    app->wsize.x = nx;
    app->wsize.y = ny;
    app->wsize.width = nw;
    app->wsize.height = nh;
}


void resize(RootApp *app, const char *g, int x, int y, int width, int height) {
    clean_window(app);
    if (app->rootpixmap) XFreePixmap(display, app->rootpixmap);
    if (app->bgpixmap)   XFreePixmap(display, app->bgpixmap);
    set_size(app, g, x, y, width, height);
    app->rootpixmap = get_root_pixmap(app);
    app->bgpixmap = copy_pixmap(app, app->rootpixmap);
}


void set_to_text_size(RootApp *app, const char *g, const char *text) {
    if (! text) return;
    resize(app, g, 0, 0, text_width(text), font_height());
}


void root_print(RootApp *app, const char *data, int x, int y) {
    HASAPP;
    XDrawString(display, app->bgpixmap, gc,
            x, y + font_ascent(),
            data, strlen(data));
}


void open_x11(const char *displayname) {
    Window window;
    int    screen;
    if ((display = XOpenDisplay(displayname)) == NULL) {
        fprintf(stderr, "Cannot open display !\n");
        exit(1);
    }
    screen = DefaultScreen(display);
    window = RootWindow(display, screen);
    gc = XCreateGC(display, window, 0, NULL);
    XSetGraphicsExposures(display, gc, False);
}


RootApp *create_app(int mask, const char *geometry) {
    RootApp *app = xmalloc(sizeof(RootApp));
    int      screen = DefaultScreen(display);

    app->window = RootWindow(display, screen);
    XSelectInput(display, app->window, mask);
    set_size(app, geometry, 0, 0, 1, 1);
    app->rootpixmap = get_root_pixmap(app);
    app->bgpixmap = copy_pixmap(app, app->rootpixmap);

    XMapWindow(display, app->window);

    return app;
}


void close_app(RootApp *app) {
    HASAPP;
    clean_window(app);
    XFreePixmap(display, app->rootpixmap);
    XFreePixmap(display, app->bgpixmap);
    XDestroyWindow(display, app->window);
    free(app);
}


void close_x11() {
    XFreeGC(display, gc);
    XCloseDisplay(display);
}


int nextevent_or_timeout(XEvent *event, unsigned long miliseconds) {
    struct timeval timeout;
    fd_set rset;

    if (XPending(display) || (miliseconds == 0)) {
        XNextEvent(display, event);
        return True;
    }

    timeout.tv_sec = miliseconds / 1000;
    timeout.tv_usec = (miliseconds % 1000) * 1000;

    FD_ZERO(&rset);
    FD_SET(ConnectionNumber(display), &rset);
    if (select(ConnectionNumber(display)+1, &rset, NULL, NULL, &timeout) > 0) {
        XNextEvent(display, event);
        return True;
    }

    return False;
}


int smooth_nextevent_or_timeout(XEvent *event, unsigned long miliseconds) {
    static struct timeval timeout;
    struct timeval        tv1, tv2;
    fd_set                rset;
    int                   sel;

    if (XPending(display) || (miliseconds == 0)) {
        XNextEvent(display, event);
        return True;
    }

    if (! timerisset(&timeout)) {
        timeout.tv_sec = miliseconds / 1000;
        timeout.tv_usec = (miliseconds % 1000) * 1000;
    }

    FD_ZERO(&rset);
    FD_SET(ConnectionNumber(display), &rset);

    gettimeofday(&tv1, NULL);
    sel = select(ConnectionNumber(display)+1, &rset, NULL, NULL, &timeout);
    gettimeofday(&tv2, NULL);

    if (sel > 0) {
        timeout.tv_sec = tv2.tv_sec - tv1.tv_sec;
        if (tv2.tv_usec < tv1.tv_usec)
            timeout.tv_usec = 999999 - tv1.tv_usec + tv2.tv_usec;
        else
            timeout.tv_usec = tv2.tv_usec - tv1.tv_usec;
        XNextEvent(display, event);
        return True;
    }

    timerclear(&timeout);

    return False;
}

